# coding: utf-8

import numpy as np 
# import cv2
import copy
import random
import Thinning
import itertools as itt

class World ():
	def __init__(self, world_shape=(50, 50), seed_pos=[], radius=1, metric='L1'):
		
		self.world_shape = world_shape
		self.radius = radius
		self.metric = metric

		self.seed_pos = seed_pos
		self.n_excited_type = len(self.seed_pos)
		
		self.stateList = ['+', '-', '.', '#']
		self.stateStr = ['excited', 'refractory', 'rest', 'bisector']
		
		labelList = [str(i+1) for i in range(self.n_excited_type)] + ['-', '.', '#']
		self.labelList = labelList
		self.label2state = dict(zip(labelList, ['+' for i in xrange(self.n_excited_type)]+self.stateList[1:]))

		self.colorList = list(itt.product((0, 32, 64, 128, 180, 223, 255),repeat=3))[1:-1]
		random.shuffle(self.colorList)
		self.label2color = dict(zip(labelList, self.colorList[:len(labelList)]))
		# print labelList
		# print ['+' for i in xrange(self.n_excited_type)] + self.stateList[1:]

		self.world = [
						[
							Cell([i, j], label='.', label2state=self.label2state) for j in xrange(world_shape[1])
						] for i in xrange(world_shape[0])
					]
		self.history = []
		self.video = []


		counter = 1
		for pos in seed_pos:
			self.world[pos[0]][pos[1]].set_label(str(counter))
			self.world[pos[0]][pos[1]].color = self.label2color[self.world[pos[0]][pos[1]].label]
			counter += 1


	def set_seed(self, seed_pos):
		self.seed_pos = seed_pos
		self.n_excited_type = len(self.seed_pos)
		labelList = [str(i+1) for i in range(self.n_excited_type)] + ['-', '.', '#']
		self.labelList = labelList
		self.label2state = dict(zip(labelList, ['+' for i in xrange(self.n_excited_type)]+self.stateList[1:]))

		self.label2color = dict(zip(labelList, self.colorList[:len(labelList)]))
		# print labelList
		# print ['+' for i in xrange(self.n_excited_type)] + self.stateList[1:]

		self.world = [
						[
							Cell([i, j], label='.', label2state=self.label2state) for j in xrange(self.world_shape[1])
						] for i in xrange(self.world_shape[0])
					]

	# def append_seed(self, seed_pos):
	# 	self.seed_pos = self.seed_pos + seed_pos
	# 	self.n_excited_type = len(self.seed_pos)
	# 	labelList = [str(i+1) for i in range(self.n_excited_type)] + ['-', '.', '#']
	# 	self.labelList = labelList
	# 	self.label2state = dict(zip(labelList, ['+' for i in xrange(self.n_excited_type)]+self.stateList[1:]))

	# 	self.label2color = dict(zip(labelList, self.colorList[:len(labelList)]))
	# 	# print labelList
	# 	# print ['+' for i in xrange(self.n_excited_type)] + self.stateList[1:]

	# 	self.world = [
	# 					[
	# 						Cell([i, j], label='.', label2state=self.label2state) for j in xrange(self.world_shape[1])
	# 					] for i in xrange(self.world_shape[0])
	# 				]

	def label_to_state(self, label):
		if label not in self.labelList:
			print 'error @ label_to_state'
			print 'wrong label: ', label, ' is used'
			assert(label in self.labelList)

		if label == '-' or label == '.' or label == '#':
			return label
		else:
			return '+'

	def dist(self, pos1, pos2):
		if self.metric == 'L1':
			return abs(pos1[0] - pos2[0]) + abs(pos1[1] - pos2[1])
		elif self.metric == 'Linf':
			return max(abs(pos1[0] - pos2[0]), abs(pos1[1] - pos2[1]))
		elif self.metric == 'L2':
			return np.sqrt((pos1[0] - pos2[0])**2 + (pos1[1] - pos2[1])**2)
		else:
			print 'error @ dist'

	def neighbors(self, pos):
		neighbor = []
		for i in xrange(max(0, pos[0]-self.radius), min(self.world_shape[0], pos[0]+self.radius+1)):
			for j in xrange(max(0, pos[1]-self.radius), min(self.world_shape[1], pos[1]+self.radius+1)):
				# if [i, j] != pos and dist(pos, [i, j], metric=metric) <= radius:
				if self.dist(pos, [i, j]) <= self.radius: # 自身を含んだ方がただしいっぽい
					neighbor.append(self.world[i][j])

		return neighbor

	def neighbor_label_set(self, pos):
		neighbor = self.neighbors(pos)
		return set([cell.label for cell in neighbor])
		
	def indicator(self, cell):
		neighbor_labels = self.neighbor_label_set(cell.pos)
		return list(neighbor_labels.intersection(set(self.labelList[:self.n_excited_type])))

	def next_label(self, cell):
		I = self.indicator(cell)
		I_state = [self.label_to_state(l) for l in I]

		if cell.state_symbol() == '+':
			if len(I) == 0 or I_state == ['+']:
				return '-'
			elif len(I) >= 2:
				return '#'

		elif cell.state_symbol() == '-':
			return '.'

		elif cell.state_symbol() == '.':
			if I == []:
				return '.'
			elif I_state == ['+']:
				return I[0]
			elif len(I) >= 2:
				return '#'

		elif cell.state_symbol() == '#':
			return '#'

		else:
			print 'error!! @ update_state'
			# print cell.state_symbol
			assert(False)
			return '_'


	def update_world(self):
		self.history.append(self.world)
		self.video.append(self.world_picture())

		changed = False
		h, w = self.world_shape
		next_world = copy.deepcopy(self.world)
		for i in xrange(h):
			for j in xrange(w):
				label = self.next_label(self.world[i][j])
				next_world[i][j].set_label(label)

				if next_world[i][j].label != self.world[i][j].label:
					changed = True


		# if self.equal(next_world):
			# return False
		
		self.world = next_world
		return changed


	def print_world(self):
		h, w = self.world_shape
		thinned_world = copy.deepcopy(self.world)
		thinned_world = Thinning.thinning(thinned_world, ['#'], ['.'])

		# print "\nRaw world"
		# for i in xrange(h):
		# 	for j in xrange(w):
		# 		if (i, j) in seed_pos:
		# 			print 'ஐ',
		# 		else:
		# 			print world[i][j].label,

		# 	print '' 

		# print "\nThinned world"
		for i in xrange(h):
			for j in xrange(w):
				if (i, j) in self.seed_pos:
					print 'ஐ',
				else:
					print thinned_world[i][j].label,

			print '' 


	# def draw_world(self, seed_pos, wait=1):
	# 	h, w = self.world_shape
	# 	frame = np.zeros((h, w, 3), np.uint8)

	# 	for i in xrange(h):
	# 		for j in xrange(w):
	# 			if self.world[i][j].label == '#' or (i, j) in seed_pos:
	# 				frame[i, j] = (0, 255, 255)
	# 			else:
	# 				frame[i, j] = (0, 0, 0)

	# 	cv2.imshow('world', frame)
	# 	cv2.waitKey(wait)
	# 	cv2.destroyAllWindows()
	# 	# if cv2.waitKey(1000) & 0xFF == ord('q'):
			# break
			
	def world_picture(self):
		h, w = self.world_shape
		frame = np.zeros((h, w, 3), np.uint8)
		thinned_world = copy.deepcopy(self.world)
		thinned_world = Thinning.thinning(thinned_world, ['#'], ['.'])

		for i in xrange(h):
			for j in xrange(w):
				if self.world[i][j].color is not None:
					frame[i][j] = self.world[i][j].color
				elif thinned_world[i][j].label == '#':
					frame[i, j] = self.label2color['#']
				elif self.world[i][j].label is not '#':
					# frame[i, j] = (0, 0, 0)
					frame[i, j] = self.label2color[self.world[i][j].label]
				else:
					frame[i, j] = self.label2color['.']

		return frame


	# def show_evolution(self):
		
	# 	for frame in self.video:
	# 		cv2.imshow('world', frame)
	# 		if cv2.waitKey(100) & 0xFF == ord('q'):
	# 			cv2.destroyAllWindows()

	# 	if cv2.waitKey(10000) & 0xFF == ord('q'):
	# 		cv2.destroyAllWindows()

	def equal(self, other_world):
		h, w = self.world_shape

		for i in xrange(h):
			for j in xrange(w):
				if self.world[i][j].label != other_world[i][j].label:
					return False

		return True
		
		
class Cell():
	"""docstring for Cell"""
	def __init__(self, pos, label='.', label2state={}):
		self.label2state = label2state
		self.label = label
		self.state = label2state[label]
		self.pos = pos
		self.color = None

	def state_symbol(self):
		return self.state

	def set_label(self, label):
		self.label = label
		self.state = self.label2state[label]





# def init_world(form=''):



if __name__ == '__main__':

	world_shape = (50, 50) # (縦幅, 横幅)

	N=10 # seed の数
	seed_pos=[(random.randint(0, world_shape[0]-1), random.randint(0, world_shape[1]-1)) for i in xrange(N)] # random.randintは端点も含む
	# print seed_pos
	# def __init__(self, world_shape=(50, 50)seed_pos=[]):
	world = World(world_shape=world_shape, seed_pos=seed_pos)

	# counter = 1
	# for pos in seed_pos:
	# 	world.world[pos[0]][pos[1]].set_label(str(counter))
	# 	counter += 1

	# seed_pos = [(4, 4), (10, 10)]
	# world[4][4].set_label('1')
	# world[14][10].set_label('2')

	T = 30
	for t in xrange(T):
		print '-------------------------------------'
		print 't = %d' % (t)
		print '-------------------------------------'
		# show_world(seed_pos)
		# world.print_world()
		world.update_world()
		
	# for frame in world.video:
	# 	# print frame.shape
	# 	cv2.imshow('world', frame)
	# 	if cv2.waitKey(100) & 0xFF == ord('q'):
	# 		cv2.destroyAllWindows()

	# if cv2.waitKey(10000) & 0xFF == ord('q'):
	# 	cv2.destroyAllWindows()

